use modules::EventLoop;
use toml::value::Table;

use std::sync::Arc;
use std::sync::mpsc::Receiver;
use transformable_channels::mpsc::ExtSender;

use Message;
use event::{Event, Envelope};

use stc::Link;
use stc::searchers::{Searcher, AggregateSearcher};
use stc::searchers::yugioh::{YugiohCard, YugiohSearcher};
use stc::searchers::mtg::{MtgCard, MtgSearcher};
use stc::searchers::mediawiki::MediawikiSearcher;

use regex::Regex;
use std::collections::BTreeMap;

pub struct AutolinkModule {}

impl AutolinkModule {
    pub fn new (_: &Table, _: &Table) -> Box<EventLoop> {
        Box::new(AutolinkModule {})
    }
}

fn print_mtg_card (card: &MtgCard, message: &Message) {
    message.reply(&format!("{}", card.image_url));

    if let Some(ref cost) = card.cost {
        message.reply(&format!("**{}** (**{}**)", card.name, cost));
    } else {
        message.reply(&format!("**{}**", card.name));
    }

    message.reply(&format!("**{}**", card.typeline));

    if let Some(ref rules) = card.rules {
        message.reply(&format!("{}", rules));        
    }

    if let Some(ref flavor) = card.flavor {
        message.reply(&format!("*{}*", flavor));
    }

    if let (&Some(ref power), &Some(ref toughness)) = (&card.power, &card.toughness) {
        message.reply(&format!("{}/{}", power, toughness));
    }
}

fn print_ygo_card (card: &YugiohCard, message: &Message) {
    if let Some(ref family) = card.family {
        message.reply(&format!("**{}** ({})", card.name, family));
    } else {
        message.reply(&format!("**{}**", card.name));
    }

    if let Some(ref level) = card.level {
        message.reply(&format!("**Level**: {}", level));
    }

    if let Some(ref subtype) = card.subtype {
        message.reply(&format!("**{} - {}**", card.card_type, subtype));
    } else {
        message.reply(&format!("**{}**", card.card_type));
    }

    message.reply(&format!("{}", card.text));

    if let (&Some(ref atk), &Some(ref def)) = (&card.atk, &card.def) {
        message.reply(&format!("{}/{}", atk, def));
    }
}

fn print_any_link (link: &Link, message: &Message) {
    message.reply(&format!("**Autolink:** {} -> {}", link.label(), link.url()));
}

impl EventLoop for AutolinkModule {
    fn run (&self, _: Box<ExtSender<Event>>, receiver: Receiver<Arc<Envelope>>) {
        let link_regex = Regex::new(r"\[\[([^\[\]]*)\]\]").expect("Invalid regex...");
        let mut searchers = AggregateSearcher::new();
        searchers.add_searcher("mtg", Box::new(MtgSearcher::new()));
        searchers.add_searcher("ygo", Box::new(YugiohSearcher::new()));
        searchers.add_searcher("pk", Box::new(MediawikiSearcher::new(String::from("https://bulbapedia.bulbagarden.net/wiki/"))));
        searchers.add_searcher("gp", Box::new(MediawikiSearcher::new(String::from("https://gammapedia.monarch-pass.net/wiki/"))));
        searchers.add_searcher("ip", Box::new(MediawikiSearcher::new(String::from("http://infinitypedia.org/wiki/"))));
        searchers.add_searcher("gcl", Box::new(MediawikiSearcher::new(String::from("https://glitchcity.info/wiki/"))));
        searchers.add_searcher("wp", Box::new(MediawikiSearcher::new(String::from("https://en.wikipedia.org/wiki/"))));

        let mut searcher_cache = BTreeMap::new();

        loop {
            match receiver.recv() {
                Ok(envelope) => {
                    if let Event::Message { ref message } = envelope.event {
                        debug!("Received message from module {:?}... {:?}", envelope.from, message.content);
                        for cap in link_regex.captures_iter(&message.content) {
                            let term = cap[1].to_owned();
                            if let &mut Some(ref mut item) = searcher_cache.entry(term.to_owned()).or_insert_with(|| searchers.exact_search(&term)) {
                                print_any_link(item, message);
                                if let Some(card) = item.downcast_ref::<MtgCard>() {
                                    print_mtg_card(card, message);
                                } else if let Some(card) = item.downcast_ref::<YugiohCard>() {
                                    print_ygo_card(card, message);
                                }
                            }
                        }
                    }
                }
                Err(error) => { error!("Error {:?}", error) }
            }
        }
    }
}
