use discord;
use discord::Discord;
use discord::model::{Event, PossibleServer};

use modules::EventLoop;
use toml::value::Table;

use event;

use std::sync::Arc;
use std::sync::mpsc::Receiver;
use transformable_channels::mpsc::ExtSender;

use std::fmt;
use std::fmt::{Debug, Formatter};

use {MessageSender, Message, User, Channel};

pub struct DiscordModule {
    token: String,
    playing: String
}

const DEFAULT_PLAYING: &'static str = "tenquestionmarks 0.0.2";

impl DiscordModule {
    pub fn new (_: &Table, configuration: &Table) -> Box<EventLoop> {
        let token = configuration.get("token")
                                 .and_then(|value| value.as_str())
                                 .unwrap_or("");

        let playing = configuration.get("playing")
                                   .and_then(|value| value.as_str())
                                   .unwrap_or(DEFAULT_PLAYING);

        Box::new(DiscordModule {
            token: String::from(token),
            playing: String::from(playing)
        })
    }
}

pub struct DiscordMessageSender {
    discord: Arc<Discord>,
    channel_id: discord::model::ChannelId
}

impl MessageSender for DiscordMessageSender {
    fn send_message (&self, message: &str) {
        debug!("Send message to channel id {:?}: {:?}", self.channel_id, message);
        match self.discord.send_message(self.channel_id, message, "", false) {
            Ok(message) => { debug!("Send message succeeded: {:?}", message.id); },
            Err(err) => { error!("Send message failed: {:?}", err) }
        }
    }
}

impl Debug for DiscordMessageSender {
    fn fmt (&self, formatter: &mut Formatter) -> fmt::Result {
        write!(formatter, "DiscordMessageSender {{ channel_id: {:?} }}", self.channel_id)
    }
}

fn make_channel_object (discord: Arc<Discord>, channel: discord::model::Channel) -> Option<Channel> {
    match channel {
        discord::model::Channel::Group(group) => {
            Some(Channel {
                name: group.name.unwrap_or_else(|| String::from("Group channel")),
                description: String::from(""),
                topic: String::from(""),
                sender: Box::new(DiscordMessageSender {
                    discord: discord,
                    channel_id: group.channel_id
                })
            })
        },
        discord::model::Channel::Private(private_channel) => {
            None
        },
        discord::model::Channel::Public(public_channel) => {
            Some(channel_from_public_channel(discord, public_channel))
        }
    }
}

fn channel_from_public_channel (discord: Arc<Discord>, channel: discord::model::PublicChannel) -> Channel {
    Channel {
        name: channel.name,
        description: String::from(""),
        topic: channel.topic.unwrap_or_else(|| String::new()),
        sender: Box::new(DiscordMessageSender {
            discord: discord,
            channel_id: channel.id
        })
    }
}

impl EventLoop for DiscordModule {
    fn run (&self, sender: Box<ExtSender<event::Event>>, _: Receiver<Arc<event::Envelope>>) {
        let discord = Arc::new(Discord::from_bot_token(&self.token[..]).expect("Discord module: Login failed"));
        let (mut connection, _) = discord.connect().expect("Discord module: Connection failed");

        info!("Playing {}", self.playing);
        connection.set_game_name(self.playing.clone());

        loop {
            let event = connection.recv_event();
            debug!("Received event: {:?}", event);

            match event {
                Ok(Event::ServerCreate(server)) => {
                    match server {
                        PossibleServer::Online(server) => {
                            info!("Joined server: {}", server.name);
                            for channel in server.channels {
                                info!(" - Joined channel: {}", channel.name);

                                match sender.send(event::Event::SelfJoin {
                                    channel: channel_from_public_channel(discord.clone(), channel)
                                }) {
                                    Err(err) => error!("Error sending selfjoin event: {:?}", err),
                                    Ok(_) => {}
                                }
                            }
                        },
                        _ => {}
                    }
                },
                Ok(Event::MessageCreate(message)) => {
                    let author = User {
                        name: message.author.name.clone(),
                        sender: Box::new(DiscordMessageSender {
                            discord: discord.clone(),
                            channel_id: message.channel_id
                        })
                    };

                    let message = Message {
                        author: author,
                        content: message.content,
                        channel: discord.get_channel(message.channel_id).ok().and_then(|channel| make_channel_object(discord.clone(), channel))
                    };

                    if let Err(err) = sender.send(event::Event::Message { message: message }) {
                        error!("Error sending message event: {:?}", err)
                    }
                }
    			Ok(_) => {}
    			Err(discord::Error::Closed(code, body)) => {
	                error!("Gateway closed on us with code {:?}: {}", code, body);
    				break
    			}
    			Err(err) => error!("Received error: {:?}", err)
            }
        }
    }
}
